#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "input.h"
#include "scanner.h"
#include "error.h"
#include "codegen.h"
#include "parser.h"
#include "symbol.h"
#include "param.h"


// Allocate Storage for a Variable
void Alloc()
{
	Next();
	if (Token != 'x')
	{
		printf ("Expected 'Variable Name'\n");
		exit(-1);
	}
	CheckDup(Value);
	AddEntry(Value, 'v');
	Allocate(Value, "0");
	Next();
}


void Block();


// Parse and Translate a Begin-Block
void BeginBlock()
{
	MatchString("BEGIN");
	Block();
	MatchString("END");
	Semi();
}




// Parse and Translate a Main Program
void DoMain()
{
	char Name[33];

	MatchString("PROGRAM");
	memcpy(Name, Value, 33);
	CheckDup(Name);
	Semi();
	Prolog();
	Next();
	BeginBlock();
}


// Process a Formal Parameter
void FormalParam()
{
	char  Name[33];

	memcpy(Name, Value, 33);
	AddParam(Name);
	Next();
}



// Process the Formal Parameter List of a Procedure
void FormalList()
{
     	MatchString("(");
     	if (Token != ')')
	{
		FormalParam();
		while (Token == ',')
		{
			Next();
               		FormalParam();
          	}
	}
	MatchString(")");
	Base = NumParams;
	NumParams = NumParams + 2;

}


// Allocate Storage on stack for a local variable
void LocalAlloc()
{
	Next();
	if (Token != 'x')
	{
		printf ("Expected 'Variable Name'\n");
		exit(-1);
	}
	AddParam(Value);
	Next();
}


// Parse and Translate a Local Data Declaration
int LocDecl()
{
	int n = 0;

	while (Scan(), Token == 'v')
	{
		LocalAlloc();
		n++;
		while(Token == ',')
		{
			LocalAlloc();
			n++;
		}
		Semi();
	}
	return (n);
}



// Parse and Translate Local Declarations 
int LocDecls()
{
	int n = 0;

	while (Scan(), Token == 'v')
	{
		n = n + LocDecl();
	}
	return(n);
}



// Parse and Translate a Procedure Declaration
void DoProc()
{
	char Name[33];
	int k;

	MatchString("PROCEDURE");
	memcpy (Name, Value, 33);
	CheckDup(Value);
	AddEntry(Value, 'p');
	Next();
	FormalList();
	k = LocDecls();
	ProcProlog(Name, k);
	BeginBlock();
	ProcEpilog();
	ClearParams();
}


// Parse and Translate a Function Declaration
void DoFunc()
{
	char Name[33];
	int k;

	MatchString("FUNCTION");
	memcpy (Name, Value, 33);
	CheckDup(Value);
	AddEntry(Value, 'p');
	Next();
	FormalList();
	k = LocDecls();
	ProcProlog(Name, k);
	BeginBlock();
	ProcEpilog();
	ClearParams();
}


// Parse and Translate a Data Declaration
void Decl()
{
	//Scan();
	while (Scan(), Token == 'v')            // FIXME IS THIS COOL OR NOT !!!!
	{
		Alloc();
		while (Token == ',')
		{
			Alloc();
		}
		Semi();
	}
}


// Parse and Translate Global Declarations
void TopDecls()
{
	Scan();
	while (Token != '.')
	{
      		switch (Token)
		{
            		case 'v':
				Decl();
				break;
			case 'p':
				DoProc();
				break;
			case 'f':
				DoFunc();
				break;
			case 'P':
				DoMain();
				break;
			default:
				printf ("TopDecls: Unregognized keyword\n");
				exit(-1);
				break;
		}
		Scan();
	}
}



#if 0
// Parse and Translate Global Declarations
void TopDecls()
{
	//Scan();
	while (Scan(), Token == 'v')            // FIXME IS THIS COOL OR NOT !!!!
	{
		Alloc();
		while (Token == ',')
		{
         		Alloc();
		}
		Semi();
	}
}
#endif

// Generate a Unique Label
char* NewLabel()
{
	char* lab;
	static int Count;

	lab = malloc (32);
	sprintf (lab, "L_%06d",  Count++);
	return (lab);
}


void BoolExpression();


// Recognize and Translate an IF Construct
void DoIf()
{

	char* L1;
	char* L2;

	Next();
	BoolExpression();
	L1 = NewLabel();
	L2 = L1;
	BranchFalse(L1);
	Block();
	if (Token == 'l')
	{
		Next();
		L2 = NewLabel();
		Branch(L2);
		PostLabel(L1);
		Block();
	}
	PostLabel(L2);
	MatchString("ENDIF");
}

// Parse and Translate a WHILE Statement
void DoWhile()
{
	char*  L1;
	char*  L2;

	Next();
	L1 = NewLabel();
	L2 = NewLabel();
	PostLabel(L1);
	BoolExpression();
	BranchFalse(L2);
	Block();
	MatchString("ENDWHILE");
	Branch(L1);
	PostLabel(L2);
}

// Check to Make Sure the Current Token is an Identifier
void CheckIdent()
{
   	if (Token != 'x')
	{
		printf ("CheckIdent: expected identifier\n");
		exit(-1);
	}
}

// Check to See if an Identifier is in the Symbol Table
// Report an error if it's not
void CheckTable(char* Symbol)
{
	if (InTable(Symbol) == 0)
	{
 		printf("CheckTable: Undefined %s\n", Symbol);
		exit(-1);
	}
}

// Read a Single Variable
void ReadVar()
{
	ReadIt();
	if (IsParam(Value))
	{
		StoreParam(ParamNumber(Value));
	}
	else
	{
		CheckTable(Value);
		StoreVariable(Value);
	}
	Next();
}


// Process a Read Statement
void DoRead()
{
	Next();
	MatchString("(");
	ReadVar();
	while (Token == ',')
	{
		Next();
      		ReadVar();
   	}
	MatchString(")");
}


// Parse and Translate a Math Factor
void Factor()
{
	if (Token == '(')
	{
		Next();
		BoolExpression();
		MatchString(")");
	}
	else
	{
		if (Token == 'x')
		{
			if (IsParam(Value))
			{
				LoadParam(ParamNumber(Value));
			}
			else
			{
				CheckTable(Value);
				LoadVariable(Value);
			}
		}
		else if (Token == '#')
		{
			LoadConstant(Value);
		}
		else
		{
			Expected("Math Factor");
		}
		Next();
	}
}

// Recognize and Translate a Multiply
void Multiply()
{
   Next();
   Factor();
   PopMul();
}

// Recognize and Translate a Divide
void Divide()
{
   Next();
   Factor();
   PopDiv();
}


// Parse and Translate a Math Term
void Term()
{
	Factor();
	while (IsMulop(Token))
	{
		Push();
		switch (Token)
		{
			case '*':
				Multiply();
				break;
       			case '/':
				Divide();
				break;
		}
	}
}

// Recognize and Translate an Add
void Add()
{
	Next();
	Term();
	PopAdd();
}


// Recognize and Translate a Subtract
void Subtract()
{
	Next();
	Term();
	PopSub();
}


// Parse and Translate an Expression
void Expression()
{
	if (IsAddop(Token))
	{
		Clear();
	}
	else
	{
		Term();
	}
	while (IsAddop(Token))
	{
		Push();
		switch (Token)
		{
			case '+':
				Add();
				break;
			case '-':
				Subtract();
				break;
		}
	}
}



// Process a Write Statement
void DoWrite()
{
	Next();
	MatchString("(");
	Expression();
	WriteIt();
	while (Token == ',')
	{
		Next();
		Expression();
		WriteIt();
	}
	MatchString(")");
}

// Get Another Expression and Compare
void CompareExpression()
{
	Expression();
	PopCompare();
}


// Get The Next Expression and Compare
void NextExpression()
{
   Next();
   CompareExpression();
}



// Recognize and Translate a Relational "Equals"
void Equal()
{
   NextExpression();
   SetEqual();
}

// Recognize and Translate a Relational "Less Than or Equal"
void LessOrEqual()
{
	NextExpression();
	SetLessOrEqual();
}

// Recognize and Translate a Relational "Not Equals"
void NotEqual()
{
   NextExpression();
   SetNEqual();
}


// Recognize and Translate a Relational "Less Than"
void Less()
{
	Next();
	switch (Token)
	{
		case '=':
			LessOrEqual();
			break;
		case '>':
			NotEqual();
			break;
		default:
			CompareExpression();
			SetLess();
        		break;
	}
}

// Recognize and Translate a Relational "Greater Than"
void Greater()
{
	Next();
	if (Token == '=')
	{
		NextExpression();
		SetGreaterOrEqual();
	}
	else
	{
		CompareExpression();
		SetGreater();
	}
}


// Parse and Translate a Relation
void Relation()
{
	Expression();
	if (IsRelop(Token))
	{
		Push();
		switch (Token)
		{
			case '=':
				Equal();
				break;
			case '<':
				Less();
				break;
			case '>':
				Greater();
				break;
		}
	}
}




// Parse and Translate a Boolean Factor with Leading NOT
void NotFactor()
{
	if (Token == '!')
	{
		Next();
		Relation();
		NotIt();
	}
	else
	{
		Relation();
	}
}


// Parse and Translate a Boolean Term
void BoolTerm()
{
	NotFactor();
	while (Token == '&')
	{
		Push();
		Next();
		NotFactor();
		PopAnd();
	}
}


// Recognize and Translate a Boolean OR
void BoolOr()
{
	Next();
	BoolTerm();
	PopOr();
}


// Recognize and Translate an Exclusive Or
void BoolXor()
{
   Next();
   BoolTerm();
   PopXor();
}


// Parse and Translate a Boolean Expression
void BoolExpression()
{
	BoolTerm();
	while (IsOrOp(Token))
	{
		Push();
		switch (Token)
		{
			case '|':
				BoolOr();
				break;
       			case '~':
				BoolXor();
				break;
		}
	}
}

// Parse and Translate an Assignment Statement 
void Assignment()
{
	char Name[33];

	//CheckTable(Value);
	memcpy (Name, Value, 33);
	Next();
	MatchString("=");
	BoolExpression();
	if (IsParam(Name))
	{
		StoreParam(ParamNumber(Name));
	}
	else
	{
		StoreVariable(Name);
	}
}


// Process an Actual Parameter
void Param()
{
	Expression();
	Push();
}


// Process the Parameter List for a Procedure  Call
int ParamList()
{
	int bytesPushed = 0;

	MatchString ("(");
	if (Token != ')')
	{
		Param();
		bytesPushed += 4;
		while (Token == ',')
		{
			Next();
			Param();
			bytesPushed += 4;
		}
	}
	MatchString (")");
	return (bytesPushed);
}


// Process a Procedure Call
void CallProc(char* Name)
{
	int bytesPushed;

	Next();
	bytesPushed = ParamList();
	Call(Name);
	CleanStack(bytesPushed);
}


// Decide if a Statement is an Assignment or Procedure Call
void AssignOrProc()
{
	char Name[33];
	memcpy(Name, Value, 33);
	switch (TypeOf(Name))
	{
		case ' ':
			printf("Identifier %s is undefined\n", Name);
			exit(-1);
			break;
		case 'v':
		case 'f':
			Assignment();
			break;
		case 'p':
			CallProc(Name);
			break;
		default:
			printf ("Identifier %s cannot be used here\n", Name);
			exit(-1);
			break;
	}
}






// Parse and Translate a Block of Statements
void Block()
{
	
   	while (Scan(), (Token != 'e') && (Token != 'l'))	// FIXME IS THIS COOL OR NOT !!!!
	{
		switch (Token)
		{
			case 'i':
				DoIf();
				break;
			case 'w':
				DoWhile();
				break;
			case 'R':
				DoRead();
				break;
			case 'W':
				DoWrite();
				break;
			case 'x':
				AssignOrProc();
				break;
		}
		Semi();
	}
}











