'----------------------------------------------------------------------------------
'Here's some code from the Spin interpreter that computes square root (Chip Gracey):

' y = 32-bit unsigned input
' x = 16-bit square root output

sqrt                    mov     x,#0
                        mov     t1,#0
                        mov     t2,#16

:loop                   shl     y,#1            wc
                        rcl     t1,#1
                        shl     y,#1            wc
                        rcl     t1,#1
                        shl     x,#2
                        or      x,#1
                        cmpsub  t1,x            wc
                        shr     x,#2
                        rcl     x,#1
                        djnz    t2,#:loop

'----------------------------------------------------------------------------------


'Faster code for  square root (Chip Gracey after discussion with lonesock on Propeller Forums):

        mov     root,#0         'reset root
        mov     mask,h40000000  'reset mask (constant in register)
:loop
        or      root,mask       'set trial bit
        cmpsub  input,root  wc  'subtract root from input if fits
        sumnc   root,mask       'cancel trial bit, set root bit if fit
        shr     root,#1         'shift root down
        shr     mask,#2         'shift mask down
        tjnz    mask,#:loop     'loop until mask empty

'This could be straight-line coded in only 4 instructions per iteration for 312,500 roots per second:

        mov     root,h40000000
        cmpsub  input,root  wc
        sumnc   root,h40000000
        shr     root,#1

        or      root,h10000000
        cmpsub  input,root  wc
        sumnc   root,h10000000
        shr     root,#1

        or      root,h04000000
        cmpsub  input,root  wc
        sumnc   root,h04000000
        shr     root,#1

        or      root,h01000000
        cmpsub  input,root  wc
        sumnc   root,h01000000
        shr     root,#1

        or      root,h00400000
        cmpsub  input,root  wc
        sumnc   root,h00400000
        shr     root,#1

        or      root,h00100000
        cmpsub  input,root  wc
        sumnc   root,h00100000
        shr     root,#1

        or      root,h00040000
        cmpsub  input,root  wc
        sumnc   root,h00040000
        shr     root,#1

        or      root,h00010000
        cmpsub  input,root  wc
        sumnc   root,h00010000
        shr     root,#1

        or      root,h00004000
        cmpsub  input,root  wc
        sumnc   root,h00004000
        shr     root,#1

        or      root,h00001000
        cmpsub  input,root  wc
        sumnc   root,h00001000
        shr     root,#1

        or      root,h00000400
        cmpsub  input,root  wc
        sumnc   root,h00000400
        shr     root,#1

        or      root,#$100
        cmpsub  input,root  wc
        sumnc   root,#$100
        shr     root,#1

        or      root,#$40
        cmpsub  input,root  wc
        sumnc   root,#$40
        shr     root,#1

        or      root,#$10
        cmpsub  input,root  wc
        sumnc   root,#$10
        shr     root,#1

        or      root,#$4
        cmpsub  input,root  wc
        sumnc   root,#$4
        shr     root,#1

        or      root,#$1
        cmpsub  input,root  wc
        sumnc   root,#$1
        shr     root,#1

'----------------------------------------------------------------------------------
